package cn.wisenergy.service.shir.filter;

import cn.wisenergy.common.utils.exception.Result;
import cn.wisenergy.model.app.AccountInfo;
import cn.wisenergy.model.app.Staff;
import cn.wisenergy.model.app.User;
import cn.wisenergy.service.shir.cache.MySessionManager;
import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.serializer.SerializerFeature;
import com.itextpdf.text.log.Logger;
import com.itextpdf.text.log.LoggerFactory;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang.exception.ExceptionUtils;
import org.apache.shiro.SecurityUtils;
import org.apache.shiro.cache.Cache;
import org.apache.shiro.cache.CacheManager;
import org.apache.shiro.session.Session;
import org.apache.shiro.session.mgt.DefaultSessionKey;
import org.apache.shiro.subject.Subject;
import org.apache.shiro.web.filter.AccessControlFilter;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.redis.core.RedisTemplate;

import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.Serializable;
import java.util.Deque;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;
import java.util.concurrent.TimeUnit;

@Slf4j
public class KickoutSessionControlFilter extends AccessControlFilter{

    private boolean kickoutAfter = false; //踢出之前登录的/之后登录的用户 默认踢出之前登录的用户
    private int maxSession = 1; //同一个帐号最大会话数 默认1

    private MySessionManager sessionManager;

    //设置session超时时长，30分钟（1800000毫秒）
    //session.setTimeout(1800000);
    //5分钟
    private static final long EXPIRE_TIME = 5 * 60;

    private static final String DEFAULT_KICKOUT_CACHE_KEY_PREFIX = "shiro:cache:kickout:";

    private String keyprefix = DEFAULT_KICKOUT_CACHE_KEY_PREFIX;

    @Autowired
    private RedisTemplate redisTemplate;

    public void setKickoutAfter(boolean kickoutAfter) {
        this.kickoutAfter = kickoutAfter;
    }

    public void setMaxSession(int maxSession) {
        this.maxSession = maxSession;
    }

    public void setSessionManager(MySessionManager sessionManager) {
        this.sessionManager = sessionManager;
    }

    @Override
    protected boolean isAccessAllowed(ServletRequest request, ServletResponse response, Object mappedValue) throws Exception {
        //如果是小程序，就放行
        boolean filter = isFilter((HttpServletRequest) request);
        return filter;
    }

    @Override
    protected boolean onAccessDenied(ServletRequest request, ServletResponse response) throws Exception {
        log.info("KickoutSessionControlFilter Not Login begin.......");
        Subject subject = getSubject(request, response);
        Map<String, Object> map = new HashMap<>();
        HttpServletResponse httpServletResponse = (HttpServletResponse) response;
        // 没有登陆或没有设置“记住我”
        HttpServletRequest request1 = (HttpServletRequest) request;
        log.info("KickoutSessionControlFilter isAuthenticated {} .......",subject.isAuthenticated());
        log.info("KickoutSessionControlFilter isRemembered {} .......",subject.isRemembered());
        log.info("KickoutSessionControlFilter  isAuthenticated isRemembered {} {} .......",request1.getRequestURI(),!subject.isAuthenticated() && !subject.isRemembered());
        if (!subject.isAuthenticated() && !subject.isRemembered()) {
            log.info("KickoutSessionControlFilter isAuthenticated and isRemembered Please login first .......");
            // 没有登陆，抛出异常
            thrLogoutException(httpServletResponse, "1001", "Please login first");
            return false;
        }
        Session session = subject.getSession();
        Serializable sessionId = session.getId();
        Integer userId = null;
        try {
            //客户端
            User user = (User) SecurityUtils.getSubject().getPrincipal();
            userId = user.getId();
        } catch (Exception e) {
            try {
                //管理端
                AccountInfo accountInfo = (AccountInfo) SecurityUtils.getSubject().getPrincipal();
                userId = accountInfo.getId();
            } catch (Exception en) {
                //员工端
                Staff staff = (Staff) SecurityUtils.getSubject().getPrincipal();
                userId = staff.getId();
            }
        }
        log.info("KickoutSessionControlFilter 如果被踢出了，直接退出，重定向到踢出后的地址-----返回1002");
        //如果被踢出了，直接退出，重定向到踢出后的地址
        if (session.getAttribute("kickout") != null) {
            log.info("------" + "踢出用户" + userId + "登录sessionId=" + sessionId + "------");
            //会话被踢出了
            try {
                //退出登录
                subject.logout();
                thrLogoutException(httpServletResponse, "1002", "您已经在其他地方登录，请重新登录。如有疑问请联系管理员！");
            } catch (Exception e) {
                e.printStackTrace();
            }
            return false;
        }
        log.info("KickoutSessionControlFilter Not Login end.......");
        return true;
    }

    public boolean isFilter(HttpServletRequest request) {
        String url =  request.getRequestURI();
        boolean result = null != request.getHeader("identity") && request.getHeader("identity").equals("miniprogram");
        boolean result1 = url.indexOf("/login")!=-1;
        return result || result1;
    }

    public void changeSession(int type){
        log.info("KickoutSessionControlFilter changeSession begin.......type : "+type);
        Subject subject = SecurityUtils.getSubject();
        Session session = subject.getSession();
        session.setTimeout(300000);
        Serializable sessionId = session.getId();
        Integer userId = null;
        if(type == 1){
            User user=(User)subject.getPrincipal();
            userId=user.getId();
        }else if(type == 2){
            AccountInfo account=(AccountInfo)subject.getPrincipal();
            userId=account.getId();
        }else{
            Staff staff=(Staff)subject.getPrincipal();
            userId=staff.getId();
        }
        //读取缓存，没有就存入
        Deque<Serializable> deque =(Deque<Serializable>) redisTemplate.opsForValue().get(getRedisKickoutKey(userId));
        //如果此用户没有session队列，也就是还没有登录过，缓存中没有，就new一个空队列，不然deque对象为空，会报空指针
        if (deque == null || deque.size()==0) {
            deque = new LinkedList<>();
        }
        //如果队列里没有此sessionId，且用户没有被踢出；放入队列
        if (!deque.contains(sessionId) && session.getAttribute("kickout") == null) {
            //将sessionId存入队列
            deque.push(sessionId);
        }
        //如果队列里的sessionId数超出最大会话数，开始踢人
        while (deque.size() > maxSession) {
            Serializable kickoutSessionId = kickoutAfter?deque.removeFirst():deque.removeLast();
            log.info("KickoutSessionControlFilter changeSession kickoutSessionId {} ......."+kickoutSessionId,kickoutSessionId);
            try {
                //获取被踢出的sessionId的session对象
                Session kickoutSession = sessionManager.getSession(new DefaultSessionKey(kickoutSessionId));
                if (kickoutSession != null) {
                    //设置会话的kickout属性表示踢出了
                    kickoutSession.setAttribute("kickout", true);
                }
            } catch (Exception e) {
            }
        }
        // 更新redis中的用户登录队列
        redisTemplate.opsForValue().set(getRedisKickoutKey(userId), deque, EXPIRE_TIME, TimeUnit.SECONDS);
        log.info("KickoutSessionControlFilter changeSession end.......");
    }

    private String getRedisKickoutKey(Integer userId) {
        return keyprefix + userId;
    }

    // 抛出未登录异常
    private void thrLogoutException(HttpServletResponse response,String ErrorCode,String ErrorMsg){
        PrintWriter writer = null;
        try {
            Result result = new Result();
            result.setResult(Result.RESULT_FLG.FAIL.getValue());
            result.setErrorCode(ErrorCode);
            result.setErrorMsg(ErrorMsg);
            response.setContentType("application/json; charset=UTF-8");
            writer = response.getWriter();
            writer.write(JSON.toJSONString(result));
            writer.flush();
        } catch (Exception e) {
            IOUtils.closeQuietly(writer);
            log.error("接口异常:{}"+ExceptionUtils.getFullStackTrace(e));
        } finally {
            if (writer != null) {
                writer.close();
            }
        }
    }


}
